#!/etc/opt/openit/perl
#
# $Id: psm.dm,v 1.17 2001/11/01 15:49:53 borgar Exp $
#
# (C) Copyright 2002 by Open iT Norge AS, All rights reserved.
#
# This software may be used or copied only with the written permission of
# Open Systems Computing AS or in accordance with the terms and conditions
# stipulated in the agreement under which this software has been supplied.
#
# This file is part of Open iT Resources.
#
#HEAD##########################################################################
#
# @NAME:
#   patch
# @AUTHOR:
#   Borgar Olsen
# @CREATED:
#   Wed Oct 16 3:47:01 MET 2002
# @DESCRIPTION:
#   Install a patch in a Open iT installation
#
#/HEAD#########################################################################

#MAIN##########################################################################
#
# @HEADING Patch installation
#
# This program, which must be inculded in with all patches, will install
# a patch. This basically consists of replacing files and restarting
# the appropriate daemons
#
# The patching is configured in the file 'patch.cfg' which is included in
# the tar file.
#
#/MAIN#########################################################################

use strict;

my $cfg = "patch.cfg";
my $status = 1;
my $reinstall = 0;

my $patch_id;

umask(0022);

# Only root can patch
if ( $< ne 0 ) {
    die "Error: Patching must be run as 'root'!\n";
}

# Verify that we can find the patch.cfg file.
if (! -e $cfg) {
    die "Error: Unable to find patch config file '$cfg'!\n";
}

# Find version of Open iT installation
my $openit_version = &get_installed_version();

# Find architecture of installation
my $arch = &get_architecture();

# Check if we have installed an 32 bit Linux installation on an AMD64 and
# should try to install x86_64 binaries if any
my $force_x86_64 = &check_force_of_x86_64();
if ( $force_x86_64 ) {
    &verify_or_patch_wrapper();
}

# Find patch ID and what Open iT verson valid for
my ($patch, $requires, @versions) = &get_patch_info();

# Verify that patch can be used for current installation
if(! match_version($openit_version, @versions)) {
    die "Error: Patch $patch is not valid for Open iT $openit_version!\n";
}

my $arg;

foreach $arg (@ARGV) {
    if ($arg eq "-reinstall") {
        $reinstall = 1;
    } else {
        print "Usage: ./patch.pl [-reinstall]\n";
        exit 0;
    }
}

# Find list of previously installed patches
my %patches = &get_installed_patches();
print "Installed patches: ", join(", ", sort keys %patches), "\n";

# Verify that patch isn't already installed
if ($patches{$patch}) {
    if ($reinstall) {
        print "Warning: Patch '$patch' is already installed, replacing it.\n";
    } else {
        die "Error: Patch '$patch' is already installed!\n";
    }
}

# Verify that required patches are installed
print "Pre-required patches: $requires\n";
my @requires = split(/,/, $requires);
foreach my $req (@requires) {
    $req =~ s/\s//g;
    if (! $patches{$req}) {
	die "Error: Pre-required patch '$req' has not been installed!\n"
    }
}

# Read file and loop through all lines
open (CFG, "patch.cfg");
binmode(CFG);
while ($status && (my $line = <CFG>)) {
    next if ( $line =~ /^\s*\#.*|^\s*$/ );  # Skip comments and empty lines
    chomp($line);

    my @elems = split(":", $line);

    # Handle directive specified in $elems[0]
    $status = &handle_operations( @elems );

    # Extra handling of x86_64 binaries on i386/i686 installations
    if ( $status && $force_x86_64
	 && $elems[0] =~ /^(add|replace)$/ && $elems[1] =~ /b/ ) {
	$elems[3] .= ".amd64";
	$status = &handle_operations( @elems ) ;
    }
}
close(CFG);

# Updated patch info for installation
&update_patch_info($patch);

# Output patch status (i.e ok / failed etc).
if (! $status) {
    die "Error: Patching failed!\n";
} else {
    print "Patch installed!\n";
}


exit 0;

sub get_installed_version {
    my $file = "/var/opt/openit/inst/package/common.info";

    open(IN, $file) || die "Open iT not installed on client!\n";
    binmode(IN);
    my $line = <IN>; # Only need first line
    close(IN);

    my $version = (split(" ", $line))[1];

    # Add missing .0's
    my $dots = $version =~ tr/././;
    for (; $dots < 3 ; $dots++) {
        $version .= ".0";
    }
    print "Patching Open iT $version\n";

    return $version;
}

sub get_patch_info {
    open(INFO, "README") || die "Unable to read README file!\n";
    binmode(INFO);
    my @lines = <INFO>;
    close(INFO);

    # Get ID
    my ($id) = grep( s/\+ID:\s*(.*)\s*/$1/, @lines);

    # Get versions
    my ($versions) = grep( s/\+VALID FOR:\s*(.*)\s*/$1/, @lines);
    my @versions = split(",", $versions);

    # Get requires
    my ($requires) = grep( s/\+REQUIRES:\s*(.*\S)\s*/$1/, @lines);

    my @requires = ();
    # REQUIRES may contain a installation version which will
    # superseed the patch requirement. Test for this
    foreach my $element( split( /,/, $requires ) ) {
	if ($element =~ /or/) {
	    my ($patch, $req_version) = (split(/ or /, $element));
	    if ( compare_version( $req_version ) > 0 ) {
		push @requires, $patch;
	    }
	} elsif ($element =~ /:/) {
	    my ($range, $patch) = (split(/ : /, $element));
	    if ( &required_for_version($patch, $range) ) {
		push @requires, $patch;
	    }
	} else {
	    push @requires, $element;
	}
    }
    $requires = join ",", @requires;

    return ($id, $requires, @versions);
}


sub get_architecture {
    my $file = "/var/opt/openit/inst/product/base.prod";

    open(IN, $file) || die "Open iT not installed on client!\n";
    binmode(IN);
    my $line = <IN>; # Only need first line
    close(IN);

    return (split(" ", $line))[0];
}

sub check_force_of_x86_64 {
    return 0 if ($arch =~ /x86_64/); # We have AMD64 installation

    return 1 if ( `uname -m` =~ /x86_64/ ); #

    return 0;
}

# Add 64 binary path to wrapper
sub verify_or_patch_wrapper {
    my ($olduid, $oldgid);
    my $wrapper = "/opt/openit/bin/wrapper";

    # Read env
    open(IN, $wrapper) || die "Unable to read Open iT binary wrapper";
    binmode(IN);
    my @openitenv = <IN>;
    close(IN);

    # Already updated?
    return if ( grep( /PROG64/, @openitenv ) );

    print "Preparing installation for 64 bit binaries\n";

    # Locate place to insert stuff
    my $ind = 0;
    while ($openitenv[$ind++] !~ /PROG=/) {}
    splice(@openitenv, $ind++, 0, "PROG64=\$\{PROG\}.amd64\n");
    while ($openitenv[$ind] !~ /OPENIT_ARCH\/bin/) { $ind++ };
    splice(@openitenv, $ind, 0,
	   'elif [ -z "$IGNORE_64BIT" -a -x $OPENIT_HOME/libexec/$OPENIT_ARCH/bin/$PROG64 ]; then',
	   "\n",
	   '    exec $OPENIT_HOME/libexec/$OPENIT_ARCH/bin/$PROG64 "$@"', "\n");

    my ($dev, $ino, $mode, $nlink, $uid, $gid) = stat("$wrapper");
    if (!defined($gid)) {
        die "Failed to stat $wrapper: $!";
    }

    $olduid = $>;
    $oldgid = $(;

    $) = $gid; # Change to the group of the file
    $> = $uid; # Change to the owner of the file

    rename $wrapper, "$wrapper.old" || warn "Failed to rename $wrapper to $wrapper.old: $!";

    if (open(OUT, ">$wrapper.$$")) {
      binmode(OUT);
      print OUT @openitenv;
      close(OUT);

      chmod(0755, "$wrapper.$$");

      rename ("$wrapper.$$", $wrapper) || warn "Failed to rename $wrapper.$$ to $wrapper: $!";
    } else {
      warn("Failed to write $wrapper.$$: $!");
      rename("$wrapper.old", $wrapper);
    }

    $> = 0; # Change uid back to root
    $) = $oldgid; # Change gid back to root
}

#
# Check if the patch is requiered for the installed version
#
sub required_for_version {
    my ($patch, $range) = @_;

    if( $range =~ m/^\s*(server|client)\s*(\S.*)$/i ) {
	$range = $2;
	my $type = lc( $1 );
	if( ! -e "/opt/openit/bin/openit-$type" ) { # affected type ?
	    return 0;      # if not, ignore...
	}
    }

    my ($lower, $upper) = split("-", $range);
    if ( compare_version( $lower ) <= 0 &&        # lower <= current ?
	 compare_version( $upper ) >= 0 ) {       # upper >= current ?
	return 1;
    }

    return 0;
}

#
# Compare versions
#
#    -1 if version < installed version
#     0 if version = installed version
#     1 if version > installed version
#
sub compare_version {
    my ($version) = @_;

    my @instparts = split('\.', $openit_version);
    my @testparts = split('\.', $version);

    # Make sure version number in 4 parts (e.g 4.9.0.0)
    while (scalar(@testparts) < 4) {
	push @testparts, 0;
    }

    my $level = 0;
    while ($level < 4) {
	my $ip = $instparts[$level];
	my $tp = $testparts[$level];

	if ( $tp < $ip ) { # Version less than installed version
	    return -1;
	} elsif ( $tp > $ip ) { # Version higher than installed version
	    return 1;
	}
	$level++;
    }

    # Version indetical to installed version
    return 0;
}


sub match_version {
    my ($rel, @vrels) = @_;

    foreach my $vrel (@vrels) {
	# Remove spaces
	$vrel =~ s/\s*(.*)\s*$/$1/;

	# Transform to match
	$vrel =~ s/\./\\./g; # . -> \.
	$vrel =~ s/x/\\d+/g;  # x -> \d+

	if ($rel =~ /^$vrel/) {
	    return 1;
	}
    }

    return 0;
}

sub get_installed_patches {
    my $dir = "/var/opt/openit/inst/patches/";
    my %patches;

    if (! -d $dir) { # If dir dosn't exist, no patche have been applied
	return ();
    }

    opendir(PATCHES, $dir) || "Unable to open patch info dir '$dir'\n";
    my @files = readdir PATCHES;

    closedir(PATCHES);
    @files = grep /^\d+$/, @files;

    foreach $patch (@files) {
	$patches{$patch} = 1,
    }

    return %patches;
}


sub handle_operations {
    my ($op, @params) = @_;

    if ($op eq "replace") {
	&add_or_replace_file(0, @params);
    } elsif($op eq "restart") {
	&restart(@params);
    } elsif($op eq "run") {
	&run_program(@params);
    } elsif($op eq "unpack") {
	&unpack_tar(@params);
    } elsif($op eq "add") {
	&add_or_replace_file(1, @params);
    } elsif($op eq "mkdir") {
	&make_dir(@params);
    } elsif($op eq "link") {
	&make_link(@params);
    } else {
	print "Error: Unknwon operation '$op' found!\n";
	return 0;
    }

    return 1;
}


sub add_or_replace_file {
    my ($is_add, $type, $source, $dest) = @_;
    my $action;

    if ($is_add) {
	$action = "Adding";
    } else {
	$action = "Replacing";
    }

    if ($dest =~ /64$/) {
	$action .= " (64bit)";
    }

    print "$action '$source'!\n";

    # Modify source and destination (for binaries)
    if ($type eq "b") {
	if ( $dest =~ /64$/ ) {              # Adding AMD64 binaries
	    $source = "files/x86_64-unknown-linux/$source";
	} else {
	    $source = "files/$arch/$source";
	}
	$dest =~ s/\$ARCH/$arch/;  # Replace
    } else {
	$source = "files/$source";
    }

    # Verify that source file exists
    if (! -e $source) {
	die "Error:: File '$source' is missing from patch!\n";
    }

    if ($is_add) {
	# Verify that destination directory exists when adding file
	my $dir = $dest;
	if ( $dest !~ m#^(/|/.*/)[^/]+$# ) {
	    print "Warning: Filepath '$dest' is invalid, skipping update!\n";
	    return;
	} else {
	    $dir = $1;
	}
	if (! -d $dir ) {
	    print "Warning: Didn't find expected directory '$dir'," .
	      " skipping update!\n";
	    return;
	}
    } else {
	# Verify that destination file exists when replacing file
	if ($dest !~ /64$/ && ! -e $dest) {
	    print "Warning: File '$dest' is missing, skipping update!\n";
	    return;
	}
    }

    &copy_file($source, $dest, !$is_add, $reinstall);
#    # Checking owner of file
#    my $SU_START = "";
#    my $SU_END = "";
#    my $owner_uid = (stat($dest))[4];
#    if ($owner_uid ) { # Not owned by root
#	my $owner = (getpwuid($owner_uid))[0];
#	$SU_START = "su $owner -c '";
#	$SU_END = "' >/dev/null 2>&1";
#    }
#
#    # Process old file and adding new
#    if( -e $dest ) { # Only rename old file if it already exists
#	rename $dest, "${dest}.old";
##	system("${SU_START}/bin/mv $dest $dest.old${SU_END}");
#    }
#    print("${SU_START}/bin/cp $source $dest${SU_END}\n");
#    system("${SU_START}/bin/cp $source $dest${SU_END}");
}


sub make_dir {
    my( $dir ) = @_;
    print "Creating directory '$dir'!\n";

    # Get the components
    my @comps = split '/+', $dir;
    $dir = "/";
    $dir .= shift @comps;
    while( -d $dir && scalar @comps ) {
	$dir .= "/" . shift @comps;
    }
    if( scalar @comps == 0 && -d $dir ) {
	return;
    }
    if( $dir =~ m#/([^/]+)$# ) {
	unshift @comps, $1;
	$dir =~ s#/[^/]+$##;
    }

    # Checking owner of file
    my $SU_START = "";
    my $SU_END = "";
    my $owner_uid = (stat($dir))[4];
    if ($owner_uid ) { # Not owned by root
	my $owner = (getpwuid($owner_uid))[0];
	$SU_START = "su $owner -c '";
	$SU_END = "' >/dev/null 2>&1";
    }

    while( scalar @comps ) {
	$dir .= "/" . shift @comps;
	system( "${SU_START}mkdir $dir${SU_END}" );
    }
}

sub make_link {
    my( $target, $name ) = @_;
    print "Creating link '$name->$target'!\n";

    if( -e $name ) {
	print "Warning: Link name '$name' exists. Skipping update!\n";
	return;
    }
    my $dir = $name;
    $dir =~ s#/[^/]+$##;

    # Checking owner of file
    my $SU_START = "";
    my $SU_END = "";
    my $owner_uid = (stat($dir))[4];
    if ($owner_uid ) { # Not owned by root
	my $owner = (getpwuid($owner_uid))[0];
	$SU_START = "su $owner -c '";
	$SU_END = "' >/dev/null 2>&1";
    }

    system( "${SU_START}ln -s $target $name${SU_END}" );
}

sub restart {
    my ($type) = @_;

    if ($type eq "client") {
	print "Restarting Open iT client\n";
	system("/opt/openit/bin/openit-client restart");
    } elsif ($type eq "server") {
	print "Restarting Open iT server\n";
	system("/opt/openit/bin/openit-server restart");
    } elsif ($type eq "webserver") {
	print "Restarting Open iT web server\n";
	system("/opt/openit/bin/openit-httpd restart");
    } elsif ($type eq "all") {
        print "Restarting Open iT all Open iT daemons\n";
        system("/opt/openit/bin/openit-all restart");
    } else {
	print "Error: Restart of unknown daemon '$type'!\n";
    }

}

sub run_program {
    my ($prog, @params) = @_;

    print "Running program '$prog'!\n";

    my $command;

    if ($prog =~ /^\//) { # Full path
	$command = "$prog " . join(" ", @params);
    } else {                   # else
	$command = "./files/$prog " . join(" ", @params);
    }

    if( system("$command") ) {
	die "Error: program '$prog' failed!";
    }
}

sub unpack_tar {
    my ($archive, $dir) = @_;

    print "Unpacking archive '$archive'\n";

    # Sanity check (no remove of /);
    if ($dir =~ q'^/$' || $dir eq "") {
	die "Illegal to remove '$dir'";
    }

    # Remove directory if it exists, then create
    if (-d $dir) {
	system("rm -rf $dir");
    }
    mkdir $dir;

    my $command = "/opt/openit/bin/tar xf files/$archive -C $dir";
    if( system("$command") ) {
	die "Unpacking of '$archive' failed!";
    }
    system("chown -R root $dir");
    system("chmod -R a+rX $dir");
}

sub update_patch_info {
    my ($id) = @_;
    my $dir = "/var/opt/openit/inst/patches/";

    if (! -d $dir) { # Create patch dir if missing
	mkdir $dir || die "Error: Unable to create patch dir '$dir'\n";
    }

    # Copy README file to patch dir

    copy_file("README", "$dir/$id", 0, 1);
}

# If replace is set to 1 it expects there to be an original from which
# to read original permissions.
# If force is set then ignore errors

sub copy_file {
    my ($from, $to, $replace, $force) = @_;
    my $buf;
    my ($dev, $ino, $mode, $nlink, $uid, $gid);
    my $statfrom;
    my ($olduid, $oldgid);
    
    if (!$force && !($to =~ m/64/) && ((!$replace) && -f "$to")) {
        die "$to is already present";
    }

#   Note: We now always take uid/gid from directory and not from
#   previous file.  This is in case its a file owned by openit
#   user in a directory owned by root.
#   if ($replace) {
#       $statfrom = $to;
#   } else {
        my @p = split(/\//, $to);
        pop(@p); # Remove the filename after full path
        $statfrom = join("/", @p);
#   }
    ($dev, $ino, $mode, $nlink, $uid, $gid) = stat($statfrom);

    if (!defined($gid)) {
        die "Failed to stat $statfrom: $!";
    }

    $olduid = $>;
    $oldgid = $(;

    $) = $gid; # Change to the group of the file
    $> = $uid; # Change to the owner of the file

    if (!open(FROM, $from)) {
        die "Can't read $from: $!";
    }
    binmode(FROM);
    if (!open(TO, ">$to.$$")) {
        die "Can't write $to: $!";
    }
    binmode(TO);

    if (!$replace) {
        $mode = 0755; # A new file so we don't copy permissions of directory
    }

    while (read(FROM, $buf, 65536)) {
        print TO $buf;
    }

    close (FROM);
    close(TO);

#   Old code, we trust in umask instead
#   if (!chmod ($mode, "$to.$$")) {
#       die "Can't chmod($mode, $to.$$): $!";
#   }

    if ($replace) { # Store the original
        if (!rename($to, "$to.old")) {
            warn "Can't rename $to to $to.old: $!";
        }
    }
    if (!rename("$to.$$", $to)) {
        die "Can't rename $to.$$ to $to: $!";
    }

    # Have to set mode
    if(!chmod ($mode, $to)) {
        rename($to, "$to.failed");
        rename("$to.old", $to);
	die "Can't set mode on $to: $!, aborting patch";
    }

#   Old code, should be redundant by $> and $) but still here due to gid bug
    if (!chown ($uid, $gid, "$to")) {
        system("chgrp $gid $to");
#       warn "Can't chown($uid, $gid, $to): $!";
    }
    $> = 0; # Change uid back to root
    $) = $oldgid; # Change gid back to root

}
